import warnings
from typing import Literal, Optional, Union

import pandas as pd
import pyarrow as pa
from pydantic import BaseModel, Field
from sqlalchemy import create_engine, text
from sqlalchemy.exc import SADeprecationWarning
from trino.auth import BasicAuthentication

from viadot.config import get_source_credentials
from viadot.sources.base import Source
from viadot.utils import get_fqn

# Triggered by trino lib
warnings.filterwarnings("ignore", category=SADeprecationWarning)


class TrinoCredentials(BaseModel):
    http_scheme: str = "https"
    host: str = "localhost"
    port: int = 443
    user: str
    password: Optional[str] = None
    catalog: str
    schema_name: Optional[str] = Field(None, alias="schema")
    verify: bool = True


class Trino(Source):
    """
    A class for interacting with Trino as a database. Currently supports only generic
    and Iceberg operations.

    Args:
        credentials (TrinoCredentials): Trino credentials.
        config_key (str, optional): The key in the viadot config holding relevant credentials.
    """

    def __init__(
        self,
        credentials: TrinoCredentials = None,
        config_key: str = None,
        *args,
        **kwargs,
    ):
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = TrinoCredentials(**raw_creds).dict(
            by_alias=True
        )  # validate the credentials

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self._con = None

        self.http_scheme = self.credentials.get("http_scheme")
        self.host = self.credentials.get("host")
        self.port = self.credentials.get("port")
        self.username = self.credentials.get("user")
        self.password = self.credentials.get("password")
        self.catalog = self.credentials.get("catalog")
        self.schema = self.credentials.get("schema")
        self.verify = self.credentials.get("verify")

    @property
    def con(self):
        if self._con is None:
            connection_string = (
                f"trino://{self.username}@{self.host}:{self.port}/{self.catalog}"
            )
            connect_args = {
                "verify": self.verify,
                "auth": BasicAuthentication(self.username, self.password),
                "http_scheme": self.http_scheme,
            }
            engine = create_engine(connection_string, connect_args=connect_args)
            return engine.connect()
        return self._con

    def get_tables(self, schema_name: str) -> list[str]:
        query = f"SHOW TABLES FROM {schema_name}"
        return self.run(query)

    def drop_table(self, table_name: str, schema_name: str = None) -> None:
        fqn = get_fqn(schema_name=schema_name, table_name=table_name)
        query = f"DROP TABLE {fqn}"

        self.logger.info(f"Dropping table {fqn}...")
        self.run(query)
        self.logger.info(f"Table {fqn} has been successfully dropped.")

    def _check_if_table_exists(self, table_name: str, schema_name: str) -> None:
        query = f"""
SELECT *
FROM INFORMATION_SCHEMA.TABLES
WHERE TABLE_SCHEMA = '{schema_name}'
AND TABLE_NAME = '{table_name}'"""
        results = self.run(query)
        return len(results) > 0

    def get_schemas(self) -> list[str]:
        query = f"SHOW SCHEMAS"
        return self.run(query)

    def _check_if_schema_exists(self, schema_name: str) -> None:
        query = f"SELECT * FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = '{schema_name}'"
        results = self.run(query)
        return bool(results)

    def drop_schema(self, schema_name: str, cascade: bool = False) -> None:
        if not self._check_if_schema_exists(schema_name):
            return

        if cascade:
            tables = self.get_tables(schema_name)
            [self.drop_table(schema_name=schema_name, table_name=t) for t in tables]

        self.logger.info(f"Dropping schema {schema_name}...")
        self.run(f"DROP SCHEMA {schema_name}")
        self.logger.info(f"Schema {schema_name} has been successfully dropped.")

    def create_iceberg_schema(self, schema_name: str, location: str) -> None:
        query = f"""
CREATE SCHEMA {schema_name}
WITH (location = '{location}')
        """
        self.logger.info(f"Creating schema {schema_name}...")
        self.run(query)
        self.logger.info(f"Schema {schema_name} has been successfully created.")

    def create_iceberg_table(
        self,
        df: pd.DataFrame,
        table_name: str,
        schema_name: str = None,
        location: str = None,
        format: Literal["PARQUET", "ORC"] = "PARQUET",
        partition_cols: list[str] = None,
        sort_by: list[str] = None,
    ) -> None:
        pa_table = pa.Table.from_pandas(df)
        columns = pa_table.schema.names
        types = [self.pyarrow_to_trino_type(str(typ)) for typ in pa_table.schema.types]
        create_table_query = self._create_table_query(
            schema_name=schema_name,
            table_name=table_name,
            columns=columns,
            types=types,
            location=location,
            format=format,
            partition_cols=partition_cols,
            sort_by_cols=sort_by,
        )
        fqn = get_fqn(schema_name=schema_name, table_name=table_name)

        self.logger.info(f"Creating table {fqn}...")
        self.run(create_table_query)
        self.logger.info(f"Table {fqn} has been successfully created.")

    def _create_table_query(
        self,
        table_name: str,
        columns: list[str],
        types: list[str],
        schema_name: str = None,
        location: str = None,
        format: Literal["PARQUET", "ORC"] = "PARQUET",
        partition_cols: list[str] = None,
        sort_by_cols: list[str] = None,
    ):
        cols_and_dtypes = ",\n\t".join(
            col + " " + dtype for col, dtype in zip(columns, types)
        )
        fqn = get_fqn(schema_name=schema_name, table_name=table_name)
        with_clause = f"format = '{format}'"

        if partition_cols:
            with_clause += ",\n\tpartitioning = ARRAY" + str(partition_cols)

        if sort_by_cols:
            with_clause += ",\n\tsorted_by = ARRAY" + str(sort_by_cols)

        if location:
            with_clause += f",\n\tlocation = '{location}'"

        query = f"""
CREATE TABLE {fqn} (
    {cols_and_dtypes}
)
WITH (
    {with_clause}
)"""

        return query

    def run(self, query: str) -> Union[list[str], None]:
        self.logger.debug("Executing query:\n" + query)

        try:
            result = self.con.execute(text(query))
            self.con.commit()
            query_keywords = ["SELECT", "SHOW", "PRAGMA", "WITH"]
            if any(query.strip().upper().startswith(word) for word in query_keywords):
                results = result.fetchall()
                return [r[0] for r in results] if results else []
        except Exception as e:
            raise ValueError(f"Failed executing query:\n{query}") from e
        finally:
            self.con.close()

    @staticmethod
    def pyarrow_to_trino_type(pyarrow_type: str) -> str:
        mapping = {
            "string": "VARCHAR",
            "large_string": "VARCHAR",
            "int8": "TINYINT",
            "int16": "SMALLINT",
            "int32": "INTEGER",
            "int64": "BIGINT",
            "float": "REAL",
            "double": "DOUBLE",
            "bool": "BOOLEAN",
            "date32[day]": "DATE",
            "timestamp[ns]": "TIMESTAMP(6)",
        }
        mapped_type = mapping.get(pyarrow_type) or "VARCHAR"
        return mapped_type

    def _check_connection(self):
        try:
            self.run("select 1")
        except Exception as e:
            raise ValueError(f"Failed to connect to Trino server at {self.host}") from e



import json

import pandas as pd

from viadot.utils import (_cast_df_cols, add_viadot_metadata_columns,
                          gen_bulk_insert_query_from_df, get_fqn,
                          handle_api_request)


def test_single_quotes_inside():
    TEST_VALUE = "a'b"
    df1 = pd.DataFrame(
        {
            "a": [
                TEST_VALUE,
            ],
            "b": ["a"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    TEST_VALUE_ESCAPED = "'a''b'"
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({TEST_VALUE_ESCAPED}, 'a')"""
    ), test_insert_query


def test_single_quotes_outside():
    TEST_VALUE = "'a'"
    df1 = pd.DataFrame(
        {
            "a": [
                TEST_VALUE,
            ],
            "b": ["b"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    TEST_VALUE_ESCAPED = "'''a'''"
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({TEST_VALUE_ESCAPED}, 'b')"""
    ), test_insert_query


def test_double_quotes_inside():
    TEST_VALUE = 'a "b"'
    df1 = pd.DataFrame(
        {
            "a": [
                TEST_VALUE,
            ],
            "b": ["c"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    TEST_VALUE_ESCAPED = """'a "b"'"""
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({TEST_VALUE_ESCAPED}, 'c')"""
    ), test_insert_query


def test_handle_api_request():
    url = "https://api.restful-api.dev/objects"
    headers = {"content-type": "application/json"}
    item = {
        "name": "test_item",
        "data": {"color": "blue", "price": 135},
    }
    payload = json.dumps(item)

    response_post = handle_api_request(
        url=url, method="POST", headers=headers, data=payload
    )
    assert response_post.ok

    item_url = f"""{url}/{response_post.json()["id"]}"""
    response_get = handle_api_request(url=item_url, method="GET", headers=headers)
    assert response_get.ok
    assert response_get.json()["data"] == item["data"]

    response_delete = handle_api_request(url=item_url, method="DELETE", headers=headers)
    assert response_delete.ok


def test_add_viadot_metadata_columns():
    class TestingClass:
        @add_viadot_metadata_columns
        def to_df(self):
            my_dict = {"AA": [1, 1], "BB": [2, 2]}
            df = pd.DataFrame(my_dict)
            return df

    testing_instance = TestingClass()
    df = testing_instance.to_df()
    assert "_viadot_source" in df.columns


def test___cast_df_cols():
    TEST_DF = pd.DataFrame(
        {
            "bool_column": [True, False, True, False],
            "datetime_column": [
                "2023-05-25 10:30:00",
                "2023-05-20 ",
                "2023-05-15 10:30",
                "2023-05-10 10:30:00+00:00 ",
            ],
            "int_column": [5, 10, 15, 20],
            "object_column": ["apple", "banana", "melon", "orange"],
        }
    )
    TEST_DF["datetime_column"] = pd.to_datetime(TEST_DF["datetime_column"])
    result_df = _cast_df_cols(
        TEST_DF, types_to_convert=["datetime", "bool", "int", "object"]
    )

    assert result_df["bool_column"].dtype == pd.Int64Dtype()
    assert result_df["datetime_column"].dtype == pd.StringDtype()
    assert result_df["int_column"].dtype == pd.Int64Dtype()
    assert result_df["object_column"].dtype == pd.StringDtype()


def test_get_fqn():
    # Test with schema name.
    fqn = get_fqn(table_name="my_table", schema_name="my_schema")
    assert fqn == "my_schema.my_table"

    # Test without schema name.
    fqn = get_fqn(table_name="my_table")
    assert fqn == "my_table"



import pandas as pd
import pyspark
import pytest

from viadot.sources import Databricks

# databricks = Databricks(env="QA")
# SCHEMA = "afraijat_test"

# source_data = [
#     {
#         "Id": "wRACnHTeuw",
#         "AccountId": 123,
#         "Name": "Scott-Merritt",
#         "FirstName": "Melody",
#         "LastName": "Cook",
#         "ContactEMail": "Melody.Cook@ScottMerritt.com",
#         "MailingCity": "Elizabethfurt",
#     },
#     {
#         "Id": "CFfTlqagNlpDu",
#         "AccountId": 456,
#         "Name": "Mann-Warren",
#         "FirstName": "Wayne",
#         "LastName": "Morrison",
#         "ContactEMail": "Wayne.Morrison@MannWarren.com",
#         "MailingCity": "Kathrynmouth",
#     },
# ]
# df = pd.DataFrame(source_data)


# def test_append():
#     append_data = [
#         {
#             "Id": "UpsertTest2",
#             "AccountId": 789,
#             "Name": "new upsert-2",
#             "FirstName": "Updated",
#             "LastName": "Carter2",
#             "ContactEMail": "Adam.Carter@TurnerBlack.com",
#             "MailingCity": "Updated!Jamesport",
#         }
#     ]

#     table = "test_append_table"

#     fqn = f"{SCHEMA}.{table}"

#     databricks.create_table_from_pandas(schema=SCHEMA, table=table, df=df)

#     append_df = pd.DataFrame(append_data)

#     did_insert = databricks.insert_into(
#         schema=SCHEMA, table=table, df=append_df, if_exists="append"
#     )
#     assert did_insert

#     expected_result = df.append(append_df)

#     result = databricks.to_df(f"SELECT * FROM {fqn}")

#     assert result.shape == expected_result.shape

#     databricks.drop_table(schema=SCHEMA, table=table)


# def test_insert_wrong_schema():
#     append_data = [
#         {
#             "Id": "UpsertTest2",
#             "AccountId": 789,
#             "Name": "new upsert-2",
#             "FirstName": "Updated",
#             "LastName": "Carter2",
#             "ContactEMail": "Adam.Carter@TurnerBlack.com",
#             "MailingCity": "Updated!Jamesport",
#             "WrongField": "Wrong field",
#         }
#     ]

#     table = "test_table"

#     append_df = pd.DataFrame(append_data)

#     with pytest.raises(pyspark.sql.utils.AnalysisException):
#         did_insert = databricks.insert_into(
#             schema=SCHEMA, table=table, df=append_df, if_exists="append"
#         )


# def test_insert_non_existent_table():
#     append_data = [
#         {
#             "Id": "UpsertTest2",
#             "AccountId": 789,
#             "Name": "new upsert-2",
#             "FirstName": "Updated",
#             "LastName": "Carter2",
#             "ContactEMail": "Adam.Carter@TurnerBlack.com",
#             "MailingCity": "Updated!Jamesport",
#             "WrongField": "Wrong field",
#         }
#     ]

#     table = "test_non_existent_table"

#     append_df = pd.DataFrame(append_data)

#     with pytest.raises(ValueError):
#         did_insert = databricks.insert_into(
#             schema=SCHEMA, table=table, df=append_df, if_exists="append"
#         )


# def test_full_refresh():
#     # Assert type and values returned after full refresh

#     table = "test_full_refresh_table"
#     fqn = f"{SCHEMA}.{table}"

#     databricks.create_table_from_pandas(schema=SCHEMA, table=table, df=df)

#     full_refresh_data = [
#         {
#             "Id": "wRACnHTeuw",
#             "AccountId": 123,
#             "Name": "Scott-Merritt",
#             "FirstName": "Melody",
#             "LastName": "Cook",
#             "ContactEMail": "Melody.Cook@ScottMerritt.com",
#             "MailingCity": "Elizabethfurt",
#         }
#     ]

#     full_refresh_df = pd.DataFrame(full_refresh_data)

#     did_insert = databricks.insert_into(
#         schema=SCHEMA, table=table, df=full_refresh_df, if_exists="replace"
#     )

#     assert did_insert

#     result = databricks.to_df(f"SELECT * FROM {fqn}")

#     assert result.shape == full_refresh_df.shape
#     databricks.drop_table(schema=SCHEMA, table=table)


# def test_upsert():
#     # Upsert and check if the data type and values are correct

#     upsert_data = [
#         {
#             "Id": "UpsertTest2",
#             "AccountId": "Updated!EHNYKjSZsiy",
#             "Name": "new upsert-2",
#             "FirstName": "Updated",
#             "LastName": "Carter2",
#             "ContactEMail": "Adam.Carter@TurnerBlack.com",
#             "MailingCity": "Updated!Jamesport",
#         }
#     ]

#     table = "test_upsert"
#     fqn = f"{SCHEMA}.{table}"
#     primary_key = "Id"

#     databricks.create_table_from_pandas(schema=SCHEMA, table=table, df=df)

#     upsert_df = pd.DataFrame(upsert_data)

#     did_insert = databricks.insert_into(
#         schema=SCHEMA,
#         table=table,
#         df=upsert_df,
#         primary_key=primary_key,
#         if_exists="update",
#     )

#     assert did_insert

#     expected_result = df.append(upsert_df)

#     result = databricks.to_df(f"SELECT * FROM {fqn}")

#     assert result.shape == expected_result.shape

#     databricks.drop_table(schema=SCHEMA, table=table)


# def test_discover_schema():
#     table = "test_table"
#     schema_result = databricks.discover_schema(schema=SCHEMA, table=table)
#     expected_schema = {
#         "Id": "string",
#         "AccountId": "bigint",
#         "Name": "string",
#         "FirstName": "string",
#         "LastName": "string",
#         "ContactEMail": "string",
#         "MailingCity": "string",
#     }

#     assert schema_result == expected_schema


# def test_rollback():
#     append_data = [
#         {
#             "Id": "UpsertTest2",
#             "AccountId": 789,
#             "Name": "new upsert-2",
#             "FirstName": "Updated",
#             "LastName": "Carter2",
#             "ContactEMail": "Adam.Carter@TurnerBlack.com",
#             "MailingCity": "Updated!Jamesport",
#         }
#     ]

#     table = "test_rollback"
#     fqn = f"{SCHEMA}.{table}"

#     databricks.create_table_from_pandas(schema=SCHEMA, table=table, df=df)

#     # Get the version of the table before applying any changes
#     version_number = databricks.get_table_version(schema=SCHEMA, table=table)

#     append_df = pd.DataFrame(append_data)

#     # Append to the table
#     did_insert = databricks.insert_into(
#         schema=SCHEMA, table=table, df=append_df, if_exists="append"
#     )
#     assert did_insert

#     # Rollback to the previous table version
#     databricks.rollback(schema=SCHEMA, table=table, version_number=version_number)
#     result = databricks.to_df(f"SELECT * FROM {fqn}")

#     assert df.shape == result.shape

#     databricks.drop_table(schema=SCHEMA, table=table)
